@@ -1,38 +1,42 @@
-/* Copyright (C) 2011, 2012, 2013 Free Software Foundation, Inc.
- * 
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 2011-2013,2018
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 
+
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
 
 #include <assert.h>
+#include <string.h>
 
-#include "libguile/_scm.h"
-#include "libguile/hash.h"
-#include "libguile/eval.h"
-#include "libguile/ports.h"
-#include "libguile/bdw-gc.h"
+#include "bdw-gc.h"
+#include "eval.h"
+#include "finalizers.h"
+#include "hash.h"
+#include "pairs.h"
+#include "ports.h"
+#include "threads.h"
+#include "weak-set.h"
 
-#include "libguile/validate.h"
-#include "libguile/weak-list.h"
-#include "libguile/weak-set.h"
+#include "weak-list.h"
 
 
 /* Weak Sets
@@ -896,13 +900,7 @@ scm_weak_set_map_to_list (SCM proc, SCM set)
 void
 scm_init_weak_set ()
 {
-#include "libguile/weak-set.x"
+#include "weak-set.x"
 
   scm_i_register_async_gc_callback (vacuum_all_weak_sets);
 }
-
-/*
-  Local Variables:
-  c-file-style: "gnu"
-  End:
-*/
