@@ -1,6 +1,6 @@
 ;;; open-coding primitive procedures
 
-;; Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014, 2015 Free Software Foundation, Inc.
+;; Copyright (C) 2009-2015, 2017-2019 Free Software Foundation, Inc.
 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -47,16 +47,19 @@
     eq? eqv? equal?
     memq memv
     = < > <= >= zero? positive? negative?
-    + * - / 1- 1+ quotient remainder modulo
+    + * - / 1- 1+ quotient remainder modulo exact->inexact
     ash logand logior logxor lognot logtest logbit?
-    sqrt abs
+    sqrt abs floor ceiling sin cos tan asin acos atan
     not
     pair? null? list? symbol? vector? string? struct? number? char? nil?
     bytevector? keyword? bitvector?
 
+    symbol->string string->symbol
+
     procedure? thunk?
 
     complex? real? rational? inf? nan? integer? exact? inexact? even? odd?
+    exact-integer?
 
     char<? char<=? char>=? char>?
 
@@ -94,7 +97,7 @@
 
     string-length string-ref string-set!
 
-    allocate-struct struct-vtable make-struct struct-ref struct-set!
+    make-struct/simple struct-vtable struct-ref struct-set!
 
     bytevector-length
 
@@ -139,7 +142,7 @@
 (define *primitive-constructors*
   ;; Primitives that return a fresh object.
   '(acons cons cons* list vector make-vector
-    allocate-struct make-struct make-struct/no-tail
+    make-struct/simple
     make-prompt-tag))
 
 (define *primitive-accessors*
@@ -168,14 +171,17 @@
     eq? eqv? equal?
     = < > <= >= zero? positive? negative?
     ash logand logior logxor lognot logtest logbit?
-    + * - / 1- 1+ sqrt abs quotient remainder modulo
+    + * - / 1- 1+ sqrt abs quotient remainder modulo exact->inexact
+    floor ceiling sin cos tan asin acos atan
     not
     pair? null? nil? list?
     symbol? variable? vector? struct? string? number? char?
     bytevector? keyword? bitvector? atomic-box?
     complex? real? rational? inf? nan? integer? exact? inexact? even? odd?
+    exact-integer?
     char<? char<=? char>=? char>?
     integer->char char->integer number->string string->number
+    symbol->string string->symbol
     struct-vtable
     length string-length vector-length bytevector-length
     ;; These all should get expanded out by expand-primitives.
@@ -194,6 +200,7 @@
     not
     pair? null? nil? list?
     symbol? variable? vector? struct? string? number? char?
+    exact-integer?
     bytevector? keyword? bitvector?
     procedure? thunk? atomic-box?
     acons cons cons* list vector))
@@ -325,58 +332,78 @@
 (define (expand-primitives x)
   (pre-order expand-primcall x))
 
-;;; I actually did spend about 10 minutes trying to redo this with
-;;; syntax-rules. Patches appreciated.
-;;;
-(define-macro (define-primitive-expander sym . clauses)
-  (define (inline-args args)
-    (let lp ((in args) (out '()))
-      (cond ((null? in) `(list ,@(reverse out)))
-            ((symbol? in) `(cons* ,@(reverse out) ,in))
-            ((pair? (car in))
-             (lp (cdr in)
-                 (cons (if (eq? (caar in) 'quote)
-                           `(make-const src ,@(cdar in))
-                           `(make-primcall src ',(caar in)
-                                           ,(inline-args (cdar in))))
-                       out)))
-            ((symbol? (car in))
-             ;; assume it's locally bound
-             (lp (cdr in) (cons (car in) out)))
-            ((self-evaluating? (car in))
-             (lp (cdr in) (cons `(make-const src ,(car in)) out)))
-            (else
-             (error "what what" (car in))))))
-  (define (consequent exp)
-    (cond
-     ((pair? exp)
-      (pmatch exp
-        ((if ,test ,then ,else)
-         `(if ,test
-              ,(consequent then)
-              ,(consequent else)))
-        (else
-         `(make-primcall src ',(car exp)
-                         ,(inline-args (cdr exp))))))
-     ((symbol? exp)
-      ;; assume locally bound
-      exp)
-     ((number? exp)
-      `(make-const src ,exp))
-     ((not exp)
-      ;; failed match
-      #f)
-     (else (error "bad consequent yall" exp))))
-  `(hashq-set! *primitive-expand-table*
-               ',sym
-               (match-lambda*
-                ,@(let lp ((in clauses) (out '()))
-                    (if (null? in)
-                        (reverse (cons '(_ #f) out))
-                        (lp (cddr in)
-                            (cons `((src . ,(car in))
-                                    ,(consequent (cadr in)))
-                                  out)))))))
+(define-syntax-rule (define-primitive-expander! sym proc)
+  (hashq-set! *primitive-expand-table* sym proc))
+
+(define-syntax primitive-expander
+  (lambda (stx)
+    (define (expand-args args)
+      (syntax-case args ()
+        (() #''())
+        ((a . b) #`(cons #,(expand-expr #'a) #,(expand-args #'b)))
+        (a (expand-expr #'a))))
+    (define (expand-expr body)
+      (syntax-case body (quote)
+        (id (identifier? #'id) #'id)
+        ((quote x) #'(make-const src 'x))
+        ((op . args) #`(make-primcall src 'op #,(expand-args #'args)))
+        (x (self-evaluating? (syntax->datum #'x)) #'(make-const src x))))
+    (define (match-clauses args+body)
+      (syntax-case args+body (if)
+        (() '())
+        ((args body . args+body)
+         (cons #`(args #,(expand-expr #'body))
+               (match-clauses #'args+body)))))
+    (syntax-case stx ()
+      ((_ args+body ...)
+       #`(lambda (src . args)
+           (match args
+             #,@(match-clauses #'(args+body ...))
+             (_ #f)))))))
+
+(define-syntax-rule (define-primitive-expander sym . clauses)
+  (define-primitive-expander! 'sym (primitive-expander . clauses)))
+
+;; Oddly, scm-error is just an explicitly 5-argument `throw'.  Weird.
+(define-primitive-expander scm-error (key who message args data)
+  (throw key who message args data))
+
+(define (escape-format-directives str)
+  (string-join (string-split str #\~) "~~"))
+
+(define-primitive-expander! 'error
+  (match-lambda*
+   ((src)
+    (make-primcall src 'throw
+                   (list (make-const src 'misc-error)
+                         (make-const src #f)
+                         (make-const src "?")
+                         (make-const src #f)
+                         (make-const src #f))))
+   ((src ($ <const> src2 (? string? message)) . args)
+    (let ((msg (string-join (cons (escape-format-directives message)
+                                  (make-list (length args) "~S")))))
+      (make-primcall src 'throw
+                     (list (make-const src 'misc-error)
+                           (make-const src #f)
+                           (make-const src2 msg)
+                           (make-primcall src 'list args)
+                           (make-const src #f)))))
+   ((src message . args)
+    (let ((msg (string-join (cons "~A" (make-list (length args) "~S")))))
+      (make-primcall src 'throw
+                     (list (make-const src 'misc-error)
+                           (make-const src #f)
+                           (make-const src msg)
+                           (make-const src "?")
+                           (make-primcall src 'list (cons message args))
+                           (make-const src #f)))))))
+
+(define-primitive-expander define! (sym val)
+  (%variable-set! (module-ensure-local-variable! (current-module) sym) val))
+
+(define-primitive-expander module-define! (mod sym val)
+  (%variable-set! (module-ensure-local-variable! mod sym) val))
 
 (define-primitive-expander zero? (x)
   (= x 0))
@@ -415,6 +442,10 @@
   (x) (/ 1 x)
   (x y z ... last) (/ (/ x y . z) last))
   
+(define-primitive-expander atan
+  (x) (atan x)
+  (x y) (atan2 x y))
+
 (define-primitive-expander logior
   () 0
   (x) (logior x 0)
@@ -427,6 +458,15 @@
   (x y) (logand x y)
   (x y z ... last) (logand (logand x y . z) last))
 
+(define-primitive-expander! 'make-vector
+  (match-lambda*
+   ((src len)
+    (make-primcall src 'make-vector (list len (make-const src *unspecified*))))
+   ((src len init)
+    (make-primcall src 'make-vector (list len init)))
+   ((src . args)
+    (make-call src (make-primitive-ref src 'make-vector) args))))
+
 (define-primitive-expander caar (x) (car (car x)))
 (define-primitive-expander cadr (x) (car (cdr x)))
 (define-primitive-expander cdar (x) (cdr (car x)))
@@ -467,13 +507,6 @@
 (define-primitive-expander call/cc (proc)
   (call-with-current-continuation proc))
 
-(define-primitive-expander make-struct (vtable tail-size . args)
-  (if (and (const? tail-size)
-           (let ((n (const-exp tail-size)))
-             (and (number? n) (exact? n) (zero? n))))
-      (make-struct/no-tail vtable . args)
-      #f))
-
 (define-primitive-expander u8vector-ref (vec i)
   (bytevector-u8-ref vec i))
 (define-primitive-expander u8vector-set! (vec i x)
@@ -546,8 +579,8 @@
                                    (make-const src #f)))))))
 
 (for-each (lambda (prim-name)
-            (hashq-set! *primitive-expand-table* prim-name
-                        (chained-comparison-expander prim-name)))
+            (define-primitive-expander! prim-name
+              (chained-comparison-expander prim-name)))
           '(< > <= >= =))
 
 (define (character-comparison-expander char< <)
@@ -560,8 +593,8 @@
 
 (for-each (match-lambda
             ((char< . <)
-             (hashq-set! *primitive-expand-table* char<
-                         (character-comparison-expander char< <))))
+             (define-primitive-expander! char<
+               (character-comparison-expander char< <))))
           '((char<? . <)
             (char>? . >)
             (char<=? . <=)
@@ -592,8 +625,8 @@
                        (make-const src #f)))
     (else #f)))
 
-(hashq-set! *primitive-expand-table* 'eqv?   (maybe-simplify-to-eq 'eqv?))
-(hashq-set! *primitive-expand-table* 'equal? (maybe-simplify-to-eq 'equal?))
+(define-primitive-expander! 'eqv?   (maybe-simplify-to-eq 'eqv?))
+(define-primitive-expander! 'equal? (maybe-simplify-to-eq 'equal?))
 
 (define (expand-chained-comparisons prim)
   (case-lambda
@@ -612,26 +645,23 @@
     (else #f)))
 
 (for-each (lambda (prim)
-            (hashq-set! *primitive-expand-table* prim
-                        (expand-chained-comparisons prim)))
+            (define-primitive-expander! prim
+              (expand-chained-comparisons prim)))
  '(< <= = >= > eq?))
 
-(hashq-set! *primitive-expand-table*
-            'call-with-prompt
-            (case-lambda
-              ((src tag thunk handler)
-               (make-prompt src #f tag thunk handler))
-              (else #f)))
+(define-primitive-expander! 'call-with-prompt
+  (case-lambda
+   ((src tag thunk handler)
+    (make-prompt src #f tag thunk handler))
+   (else #f)))
 
-(hashq-set! *primitive-expand-table*
-            'abort-to-prompt*
-            (case-lambda
-              ((src tag tail-args)
-               (make-abort src tag '() tail-args))
-              (else #f)))
-(hashq-set! *primitive-expand-table*
-            'abort-to-prompt
-            (case-lambda
-              ((src tag . args)
-               (make-abort src tag args (make-const #f '())))
-              (else #f)))
+(define-primitive-expander! 'abort-to-prompt*
+  (case-lambda
+   ((src tag tail-args)
+    (make-abort src tag '() tail-args))
+   (else #f)))
+(define-primitive-expander! 'abort-to-prompt
+  (case-lambda
+   ((src tag . args)
+    (make-abort src tag args (make-const #f '())))
+   (else #f)))
